# Generated by Django 2.0 on 2017-12-18 09:39

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("codenerix", "0020_remotelog"),
    ]

    operations = [
        migrations.AddField(
            model_name="log",
            name="username",
            field=models.CharField(
                blank=True,
                default="",
                max_length=200,
                verbose_name="Username",
            ),
        ),
        migrations.AddField(
            model_name="remotelog",
            name="username",
            field=models.CharField(
                blank=True,
                default="",
                max_length=200,
                verbose_name="Username",
            ),
        ),
        migrations.AlterField(
            model_name="remotelog",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



#
# django-codenerix
#
# Codenerix GNU
#
# Project URL : http://www.codenerix.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

try:
    from subprocess import getstatusoutput

    pythoncmd = "python3"
except Exception:
    from commands import (  # type: ignore[import-not-found,no-redef]
        getstatusoutput,
    )

    pythoncmd = "python2"

from codenerix_lib.debugger import Debugger
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand, Debugger):
    # Show this when the user types help
    help = "Do a touch to the project"

    def add_arguments(self, parser):
        # Named (optional) arguments
        parser.add_argument(
            "-f",
            action="store_true",
            dest="f",
            default=False,
            help="Keep the command working forever",
        )

        # Named (optional) arguments
        parser.add_argument(
            "--follow",
            action="store_true",
            dest="follow",
            default=False,
            help="Keep the command working forever",
        )

    def handle(self, *args, **options):
        # Autoconfigure Debugger
        self.set_name("CODENERIX")
        self.set_debug()

        # Get environment
        appname = settings.ROOT_URLCONF.split(".")[0]
        basedir = settings.BASE_DIR
        appdir = os.path.abspath("{}/{}".format(basedir, appname))

        # While keep working
        keepworking = True
        while keepworking:
            # Collect
            self.debug("Collecting...", color="blue")
            cmd = f"{appdir}/manage collectstatic --noinput"
            status, output = getstatusoutput(cmd)
            if status:
                raise CommandError(f"{output}\nCommand was: {cmd}")
            for line in output.split("\n"):
                if line[0:7] == "Copying":
                    path = line.split("'")[1].replace(appdir, ".")
                    self.debug("    > {}".format(path))
                elif ("static file copied to" in line) or (
                    "static files copied to" in line
                ):
                    done = line.split(" ")[0]
                    linesp = line.split(",")
                    if len(linesp) == 1:
                        total = linesp[0].split(" ")[0]
                    else:
                        total = linesp[1].split(" ")[1]
                    self.debug(
                        f"{done}/{total} files copied",
                        color="cyan",
                    )
                elif (
                    "was already registered. Reloading models is not advised "
                    "as it can lead to inconsistencies, most notably with "
                    "related models." in line
                ):
                    pass
                elif (
                    "new_class._meta.apps.register_model(new_class._meta.app_label, new_class)"  # noqa: E501
                    in line
                ):
                    pass
                elif line == "":
                    pass
                else:
                    self.warning("Unknown string: #{}#".format(line))

            # Clean
            self.debug("Cleaning...", color="blue")
            cmd = f"{appdir}/manage clean"
            status, output = getstatusoutput(cmd)
            if status:
                raise CommandError(f"{output}\nCommand was: {cmd}")

            # Touch
            self.debug("Touch...", color="blue", tail=False)
            filenames = os.listdir(appdir)
            filenames.sort()
            for name in filenames:
                if name[0:4] == "wsgi" or name[-4:] == "wsgi":
                    cmd = f"/usr/bin/touch {appdir}/{name}"
                    status, output = getstatusoutput(cmd)
                    if status:
                        raise CommandError(f"{output}\nCommand was: {cmd}")
                    self.debug(
                        " [{}]".format(name),
                        color="cyan",
                        header=False,
                        tail=False,
                    )
            self.debug(" Done", color="green", header=False)

            if options["follow"] or options["f"]:
                # Ask the user if would really like to remove all
                # locale folders
                self.debug(
                    "Hit a ENTER when ready to go or Q to exit (ENTER|q) ",
                    tail=False,
                    color="purple",
                )
                try:
                    key = raw_input().lower()
                except NameError:
                    key = input().lower()
                except KeyboardInterrupt:
                    self.debug(" ", header=False)
                    key = "q"
                self.debug("", header=False)
                self.debug(" ")
                if key == "q":
                    self.debug("Quitting!", color="yellow")
                    keepworking = False
            else:
                self.debug("Quitting!", color="yellow")
                keepworking = False



# from django.conf import settings
from haystack.backends.elasticsearch2_backend import (  # type: ignore[import-not-found] # noqa: E501
    Elasticsearch2SearchBackend,
    Elasticsearch2SearchEngine,
)


class AsciifoldingElasticBackend(Elasticsearch2SearchBackend):
    """
    Mounir Messelmeni - 13/December/2015
    https://mounirmesselmeni.github.io/2015/12/13/enable-asciifolding-in-elasticsearchhaystack/
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        analyzer = {
            "ascii_analyser": {
                "tokenizer": "standard",
                "filter": ["standard", "asciifolding", "lowercase"],
            },
            "ngram_analyzer": {
                "type": "custom",
                "tokenizer": "lowercase",
                "filter": ["haystack_ngram", "asciifolding"],
            },
            "edgengram_analyzer": {
                "type": "custom",
                "tokenizer": "lowercase",
                "filter": ["haystack_edgengram", "asciifolding"],
            },
        }
        self.DEFAULT_SETTINGS["settings"]["analysis"]["analyzer"] = analyzer

    def build_schema(self, fields):
        content_field_name, mapping = super().build_schema(fields)

        for field_name, field_class in fields.items():
            field_mapping = mapping[field_class.index_fieldname]

            if field_mapping["type"] == "string" and field_class.indexed:
                if not hasattr(
                    field_class,
                    "facet_for",
                ) and field_class.field_type not in ("ngram", "edge_ngram"):
                    field_mapping["analyzer"] = "ascii_analyser"

            mapping.update({field_class.index_fieldname: field_mapping})
        return (content_field_name, mapping)


class AsciifoldingElasticSearchEngine(Elasticsearch2SearchEngine):
    """
    Mounir Messelmeni - 13/December/2015
    https://mounirmesselmeni.github.io/2015/12/13/enable-asciifolding-in-elasticsearchhaystack/
    """

    backend = AsciifoldingElasticBackend



# Generated by Django 5.0.6 on 2024-06-05 08:12

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("codenerix", "0022_alter_log_options"),
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterField(
            model_name="log",
            name="content_type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="log",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AlterField(
            model_name="log",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="log",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="remotelog",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="remotelog",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



#
# django-codenerix
#
# Codenerix GNU
#
# Project URL : http://www.codenerix.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from threading import local

from django.conf import settings
from django.shortcuts import redirect


class SecureRequiredMiddleware:
    def __init__(self, get_response=None):
        self.get_response = get_response
        self.paths = getattr(
            settings,
            "HTTPS_PATHS",
            getattr(settings, "SECURE_REQUIRED_PATHS", ("/",)),
        )
        self.enabled = self.paths and getattr(settings, "HTTPS_SUPPORT", True)

    def process_request(self, request):
        if self.enabled and not request.is_secure():
            # Check for negative matching, URLs we don't want to be secure
            for path in self.paths:
                full_path = request.get_full_path()
                if path[0] == "-" and (
                    full_path.startswith(path[1:])
                    or re.compile(path[1:]).match(full_path)
                ):
                    return None

            # Check for positive matching, URLs we want to be secure
            for path in self.paths:
                full_path = request.get_full_path()
                if full_path.startswith(path) or re.compile(path).match(
                    full_path,
                ):
                    request_url = request.build_absolute_uri(
                        request.get_full_path(),
                    )
                    secure_url = request_url.replace("http://", "https://")
                    return redirect(secure_url, permanent=True)

        # Any other lateral case
        return None

    def __call__(self, request):
        # Code to be executed for each request before the view (and later
        # middleware) are called.
        response = self.process_request(request)

        # Get response
        if response is None:
            response = self.get_response(request)

        # Code to be executed for each request/response after the
        # view is called
        # ... pass ...

        # Return response
        return response


class CurrentUserMiddleware:
    """
    Let the system to have the username everywhere
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def process_request(self, request):
        _user.value = request.user

    def __call__(self, request):
        # Code to be executed for each request before the view (and
        # later middleware) are called.
        self.process_request(request)

        # Get response
        response = self.get_response(request)

        # Code to be executed for each request/response after the
        # view is called
        # ... pass ...

        # Return response
        return response


def get_current_user():
    """
    Thought this function you can get the user loged in everywhere
    """
    if "value" in dir(_user):
        return _user.value
    else:
        return None


_user = local()



# Generated by Django 1.10.7 on 2017-07-25 16:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("codenerix", "0018_log_snapshot_txt"),
    ]

    operations = [
        migrations.AlterField(
            model_name="log",
            name="snapshot_txt",
            field=models.TextField(blank=True, verbose_name="Snapshot Txt"),
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-27 11:24

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("codenerix", "0021_auto_20171218_1039"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="log",
            options={
                "permissions": [
                    ("list_log", "Can list log"),
                    ("detail_log", "Can view log"),
                ],
            },
        ),
    ]



#
# django-codenerix
#
# Codenerix GNU
#
# Project URL : http://www.codenerix.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import hashlib
import imghdr
import io
import json
import os
import random
import re
from datetime import datetime

from captcha.widgets import ReCaptcha  # type: ignore[import-not-found]
from django import forms
from django.conf import settings
from django.core.files.base import File
from django.core.serializers.json import DjangoJSONEncoder
from django.urls import resolve, reverse
from django.utils import formats
from django.utils.encoding import smart_str
from django.utils.html import escapejs
from django.utils.translation import get_language
from django.utils.translation import gettext as _

from codenerix.helpers import get_class

try:
    from django.utils.choices import BlankChoiceIterator
except ModuleNotFoundError:
    # Django 4.x compatibility
    BlankChoiceIterator = list


class StaticSelectMulti(forms.widgets.SelectMultiple):
    __language = None

    def __init__(self, *args, **kwargs):
        targs = None
        if args:
            targs = args[0]
        if kwargs:
            targs = kwargs
        if targs:
            # common
            self.is_required = targs.get("is_required", False)
            self.form_name = targs.get("form_name", "")
            self.field_name = targs.get("field_name", "")
            # static
            self.choices = targs.get("choices", "")
            # dynamic
            self.autofill_deepness = targs.get("autofill_deepness", 2)
            self.autofill_url = targs.get("autofill_url", "")
            self.autofill = targs.get("autofill_related", [])
        return super().__init__(*args, **kwargs)

    def set_language(self, language):
        self.__language = language

    def value_from_datadict(self, data, files, name):
        # Get value
        value = super().value_from_datadict(data, files, name)

        # Decompress list(list) to list
        if (
            isinstance(value, list)
            and len(value) == 1
            and isinstance(value[0], list)
        ):
            value = value[0]

        # Let Django do its work
        return value

    def render(self, name, value, attrs=None, choices=(), renderer=None):
        # Initialization
        # required=self.attrs.get('ng-required','false')
        vmodel = self.field_name
        # vid=attrs.get('id','id_{0}'.format(name))
        # vstyle=attrs.get('style','')
        vform = self.form_name
        placeholder = escapejs(_("Press * and select an option"))
        if value is None:
            value = []
        valuejs = []

        # Prepare dynamic search
        if self.dynamic:
            # Check if autofill_url is defined
            if not self.autofill_url:
                raise OSError("autofill_url not defined")

            # Prepare link
            vurl = reverse(self.autofill_url, kwargs={"search": "a"})[:-1]
            # Get access to the get_label() method and request for the
            # label of the bound input
            if value:
                func = resolve(vurl + "*").func
                clss = get_class(func)
                if self.__language:
                    clss.language = self.__language
                # label = clss().get_label(value)

            # Prepare foreign params
            foreign_params = "http,'{0}',amc_items,{{".format(vurl)
            comma = False
            for field in self.autofill:
                if ":" in field:
                    field_filter = field.split(":")[-1]
                    field = field.split(":")[0]
                else:
                    field_filter = field
                if comma:
                    foreign_params += ","
                else:
                    comma = True
                foreign_params += "'{}':{}.{}".format(
                    field,
                    vform,
                    field_filter,
                )
            foreign_params += (
                "}},amc_select.{0},amc_searchText.{0},{1}".format(
                    vmodel,
                    self.autofill_deepness,
                )
            )
        else:
            foreign_params = None

        # Prepare initialization string
        init = ""
        if not value:
            value = []
        if self.static and hasattr(self, "field"):
            for key, label in self.field.choices:
                if value == key:
                    valuejs.append(
                        '{{"id":"{0}","label":"{1}"}},'.format(
                            key,
                            escapejs(smart_str(label)),
                        ),
                    )
                init += '{{"id":"{0}","label":"{1}"}},'.format(
                    key,
                    escapejs(smart_str(label)),
                )
        elif (isinstance(self.choices, BlankChoiceIterator)) or (
            isinstance(self.choices, list)
        ):
            for key, label in self.choices:
                if value == key:
                    valuejs.append(
                        '{{"id":"{0}","label":"{1}"}},'.format(
                            key,
                            escapejs(smart_str(label)),
                        ),
                    )
                init += '{{"id":"{0}","label":"{1}"}},'.format(
                    key,
                    escapejs(smart_str(label)),
                )
        else:
            # FORCE RELOAD DATAS
            elements = self.choices.queryset
            elements._result_cache = None
            if self.dynamic:
                if value and isinstance(value, list):
                    elements = elements.filter(pk__in=value)
                else:
                    elements = []
            for choice in elements:
                init += '{{"id":"{0}","label":"{1}"}},'.format(
                    choice.pk,
                    escapejs(smart_str(choice)),
                )
                if value and isinstance(value, list) and (choice.pk in value):
                    valuejs.append(
                        '{{"id":"{0}","label":"{1}"}},'.format(
                            int(choice.pk),
                            escapejs(smart_str(choice)),
                        ),
                    )

        # Build HTML
        html = (
            '<md-chips ng-model="amc_select.{0}" '
            'md-autocomplete-snap id="{0}" name="{0}" '.format(vmodel)
        )
        html += '    md-transform-chip="amc_transformChip($chip)"'
        # html += u'    initial = "loadVegetables()"'
        html += "    ng-init = 'amc_items.{} = ".format(vmodel)
        html += "[" + init + "]"
        if valuejs:
            # html += u" ng-init = 'amc_select.{0} = [".format(vmodel)
            html += "; amc_select.{} = [".format(vmodel)
            html += "".join(valuejs)
            html += "]'"
        html += "'"
        html += '    md-require-match="amc_autocompleteDemoRequireMatch">'
        html += "    <md-autocomplete"
        html += '            md-selected-item="amc_selectedItem.{}"'.format(
            vmodel,
        )
        html += '            md-search-text="amc_searchText.{}"'.format(vmodel)
        html += (
            '            md-items="item in amc_querySearch('
            "amc_searchText.{0}, '{0}'".format(vmodel)
        )

        # Add the params for dynamic search
        if self.dynamic:
            html += "," + foreign_params

        html += ')"'
        html += '            md-item-text="item.id"'
        html += '            placeholder="{}">'.format(placeholder)
        html += '        <span md-highlight-text="amc_searchText.{}">'.format(
            vmodel,
        )
        html += "            {{item.label}}</span>"
        html += "    </md-autocomplete>"
        html += "    <md-chip-template>"
        html += "        <span>"
        html += "        {{$chip.label}}"
        html += "        </span>"
        html += "    </md-chip-template>"
        html += "</md-chips>"
        return html


class StaticSelect(forms.widgets.Select):
    """
    Howto:
    1) The form class must inherit from GenModelForm
    """

    def __init__(self, *args, **kwargs):
        targs = None
        if args:
            targs = args[0]
        if kwargs:
            targs = kwargs
        if targs:
            self.is_required = targs.get("is_required", False)
            self.form_name = targs.get("form_name", "")
            self.field_name = targs.get("field_name", "")
            self.choices = targs.get("choices", "")
        return super().__init__(*args, **kwargs)

    def render(self, name, value, attrs=None, choices=(), renderer=None):
        # Initialization
        required = self.attrs.get("ng-required", "false")
        controller = self.attrs.get("ng-controller", None)
        disabled = self.attrs.get("ng-disabled", None)
        change = (
            self.attrs.get("ng-change", "undefined")
            .replace('"', "'")
            .replace("'", "\\'")
        )
        if change != "undefined":
            change = "'{}'".format(change)
        vmodel = self.field_name
        vid = attrs.get("id", "id_{}".format(name))
        vstyle = attrs.get("style", "")
        ngreadonly = attrs.get("ng-readonly", "")
        vform = smart_str(self.form_name)
        placeholder = escapejs(_("Select an option"))
        is_multiple = hasattr(self, "multiple") and self.multiple

        # Render
        html = (
            "<input name='{0}' ng-required='{1}' ng-model='{0}' "
            "type='hidden'".format(vmodel, required)
        )
        if controller:
            html += " ng-controller='{}' ".format(controller)
        if disabled:
            html += " ng-disabled='{}' ".format(disabled)
        valuejs = None
        if is_multiple:
            if value:
                if isinstance(value, list):
                    valuejs = json.dumps(
                        [int(x) for x in value],
                        cls=DjangoJSONEncoder,
                    )
                else:
                    valuejs = json.dumps(
                        [
                            int(value),
                        ],
                        cls=DjangoJSONEncoder,
                    )

                html += ' ng-init="{}={}"'.format(vmodel, valuejs)
            else:
                html += ' ng-init="{}=[]"'.format(vmodel)
        elif value:
            html += " ng-init=\"{}='{}'\"".format(vmodel, value)

        html += ">"

        html += "<ui-select "
        if disabled:
            html += " ng-disabled='{}' ".format(disabled)
        else:
            html += ' ng-disabled="{}"'.format(ngreadonly)
        if controller:
            html += " ng-controller='{}' ".format(controller)
        if is_multiple:
            html += " multiple "
        # Build options
        html += ' ng-init="options.{}=['.format(vmodel)
        ini = None
        list_value = []
        index = 0

        if is_multiple and not value:
            value = []
        if hasattr(self, "field"):
            for key, label in self.field.choices:
                if value == key:
                    ini = index
                html += "{{'id':'{0}','label':'{1}'}},".format(
                    key,
                    escapejs(smart_str(label)),
                )
                index += 1
        elif (isinstance(self.choices, BlankChoiceIterator)) or (
            isinstance(self.choices, list)
        ):
            for key, label in self.choices:
                if value == key:
                    ini = index
                html += "{{'id':'{0}','label':'{1}'}},".format(
                    key,
                    escapejs(smart_str(label)),
                )
                index += 1
        else:
            if not is_multiple:
                html += "{{'id':null, 'label':'{0}'}}, ".format(placeholder)

            for choice in self.choices.queryset:
                if value == choice.pk:  # only not multiple
                    ini = index
                html += "{{'id':'{0}','label':'{1}'}},".format(
                    choice.pk,
                    escapejs(smart_str(choice)),
                )
                if (
                    value
                    and isinstance(value, list)
                    and (str(choice.pk) in value)
                ):
                    list_value.append(
                        {"id": int(choice.pk), "label": str(choice)},
                    )
                index += 1

            # raise Exception(html)
        html += "];"
        # Select if we have already one
        if ini is not None:
            if is_multiple:
                html += "$select.selected=options.{};".format(vmodel)
            else:
                html += "$select.placeholder='options.{}[{}];'".format(
                    vmodel,
                    ini,
                )
            # ok html += u"$select.selected=options.{0}".format(vmodel,ini)

        if is_multiple and value is not None:
            html += "{}.{}.$setViewValue({});".format(
                vform,
                vmodel,
                json.dumps(list_value, cls=DjangoJSONEncoder).replace(
                    '"',
                    "'",
                ),
            )
            # html += u"$parent.{1}.$setViewValue({2});".format(vform,
            # vmodel,json.dumps(
            #     list_value,
            #     cls=DjangoJSONEncoder
            # ).replace('"', "'"))

        html += ' "'
        # Build the rest of the tag
        html += ' id="{}"'.format(vid)
        html += ' ng-model="$parent.{}"'.format(vmodel)
        if is_multiple:
            html += (
                " on-select=\"selectedOptionSelect({}.{},'{}',{}, "
                '$parent, $select.selected)"'.format(
                    vform,
                    vmodel,
                    valuejs,
                    change,
                )
            )
        else:
            if value is None:
                value = "null"
            else:
                value = "'{}'".format(value)
            html += (
                ' on-select="selectedOptionSelect({}.{},{},{}, '
                '$parent, $select.selected)"'.format(
                    vform,
                    vmodel,
                    value,
                    change,
                )
            )
        html += ' theme="bootstrap"'
        html += ' ng-disabled="disabled"'
        html += ' reset-search-input="false"'
        html += ">"
        html += ' <ui-select-match placeholder="{}">'.format(placeholder)
        if is_multiple:
            html += "    {{$item.label}}"
        else:
            html += "    {{$select.selected.label}}"
        html += " </ui-select-match>"
        html += " <ui-select-choices"
        html += '     style="{}"'.format(vstyle)
        html += '     repeat="value.id as value in options.{}'.format(vmodel)
        html += '            | filter: {label: $select.search}">'
        html += (
            '     <div ng-bind-html="value.label| highlightSelect: '
            '$select.search"></div>'
        )
        html += " </ui-select-choices>"
        html += "</ui-select>"
        # Return result
        return html


class DynamicSelectInputWidget:
    """
    Howto:

    1) The form class must inherit from GenModelForm

    2) In the form class add inside the Meta class a new attribute named
       autofill:
        autofill = {'ng-model':['select', length, baseurl, field1, field2,
                     field3, ...]}

        Meaning:
        - ng-model: model name in AngularJS
        - 'select': it is constant string 'select'
        - length: how many characters the user has to add in the input so a
          search will happen
        - baseurl: reverse url name to reach to the class that we will ask for
          the options to render
        - field1, field2, .... : other ng-model fields that are linked to this
          one and use to optimize the option's search

    3) You must define an url with a reverse name that we will use to
       configure baseurl

    4) The target (view) for this url (reverse name) must
       implement GenForeignKey.

    5) If you use the templatetag 'addattr' to add a 'ng-change' remember you
       have to operate with $externalScope which referers to the Scope outside
       the widget, this is usefull when the widget is inside some ng-repeat or
       another artificial structure generated by the programmer. To operate
       with form fields you don't need to operate with $externalScope, just
       point to the form_name.field_name as usually.

    6) In the external fields to be included with your search, you can include
       the tag <FORMNAME> and it will be dynamically replaced with the name of
       the form so you can define external fields like '$scope.var1' and/or
       '<FORMNAME>.field2'
    """

    language = None

    def __init__(self, *args, **kwargs):
        targs = None

        if args:
            targs = args[0]

        if kwargs:
            targs = kwargs

        if targs:
            self.is_required = targs.get("is_required", False)
            self.form_name = targs.get("form_name", "")
            self.field_name = targs.get("field_name", "")
            self.autofill_deepness = targs.get("autofill_deepness", 2)
            self.autofill_url = targs.get("autofill_url", "")
            self.autofill = targs.get("autofill_related", [])

        return super().__init__(*args, **kwargs)

    def set_language(self, language):
        self.language = language

    def get_foreign(self, vurl, vform, vmodel, search, function):
        html = "{0}(http,'{1}',options,{{".format(function, vurl)
        comma = False
        for field in self.autofill:
            if ":" in field:
                field_filter = field.split(":")[-1]
                field = field.split(":")[0]
            else:
                field_filter = field
            if comma:
                html += ","
            else:
                comma = True
            html += "'{}':{}".format(
                field,
                field_filter.replace("<FORMNAME>", vform),
            )
        html += "}}, '{0}', {1}, {2}, {3});\"".format(
            vmodel,
            vmodel,
            search,
            self.autofill_deepness,
        )
        return html


class DynamicSelect(DynamicSelectInputWidget, forms.widgets.Select):
    def render(self, name, value, attrs=None, choices=(), renderer=None):
        if not self.autofill_url:
            raise OSError("autofill_url not defined")
        # Initialization
        controller = self.attrs.get("ng-controller", None)
        disabled = self.attrs.get("ng-disabled", None)
        required = self.attrs.get("ng-required", "false")
        change = (
            self.attrs.get("ng-change", "undefined")
            .replace('"', "'")
            .replace("'", "\\'")
        )
        if change != "undefined":
            change = "'{}'".format(change)
        vmodel = self.field_name
        vid = attrs.get("id", "id_{}".format(name))
        vstyle = attrs.get("style", "")
        vform = smart_str(self.form_name)
        vurl = reverse(self.autofill_url, kwargs={"search": "a"})[:-1]
        # Get access to the get_label() method and request for the label of
        # the bound input
        if value:
            func = resolve(vurl + "*").func
            clss = get_class(func)
            if self.language:
                clss.language = self.language
            label = clss().get_label(value)

        # Prepare placeholder
        placeholder = self.attrs.get(
            "placeholder",
            _("Press * or start typing"),
        )
        ngplaceholder = self.attrs.get(
            "ng-placeholder",
            "'{}'".format(_("Press * or start typing")),
        )

        # Detect parent node
        vmodelsp = vmodel.split(".")
        if len(vmodelsp) > 1:
            vparent = ".".join(vmodelsp[:-1])
        else:
            vparent = "$parent"

        # Render
        html = (
            "<input name='{0}' ng-required='{1}' "
            "ng-model='{0}' id='{0}' type='hidden'".format(vmodel, required)
        )
        if controller:
            html += " ng-controller='{}' ".format(controller)
        if disabled:
            html += " ng-disabled='{}' ".format(disabled)
        # Set initial value
        if value:
            html += " ng-init=\"{}='{}'\"".format(vmodel, value)
        html += ">"
        html += "<ui-select"
        if controller:
            html += " ng-controller='{}' ".format(controller)
        if disabled:
            html += " ng-disabled='{}' ".format(disabled)
        if hasattr(self, "multiple") and self.multiple:
            html += " multiple "
        if value:
            html += (
                " ng-init=\"options.{0}=[{{'id':null,'label':{3}}},"
                "{{'id':'{1}','label':'{2}'}}]; "
                "$select.selected=options.{0}[1];".format(
                    vmodel,
                    value,
                    escapejs(label),
                    ngplaceholder,
                )
            )
            html += (
                " option_default={}; options.{}=option_default['rows']".format(
                    self.get_foreign(
                        vurl,
                        vform,
                        vmodel,
                        "'*'",
                        "getForeignKeys",
                    ),
                    vmodel,
                )
            )
        else:
            # init options for form modal
            html += (
                " ng-init=\"options.{0}=[{{'id':null,"
                "'label':{1}}}]; \"".format(vmodel, ngplaceholder)
            )
        html += (
            ' ng-click="option_default={}; '
            "options.{}=option_default['rows']".format(
                self.get_foreign(vurl, vform, vmodel, "'*'", "getForeignKeys"),
                vmodel,
            )
        )

        html += ' id="{}"'.format(vid)
        html += ' ng-model="{}.{}"'.format(vparent, vmodel)
        html += (
            " on-select=\"selectedOptionSelect({}.{},'{}',{}, "
            '$parent, $select.selected)"'.format(vform, vmodel, value, change)
        )
        html += ' theme="bootstrap"'
        html += ' ng-disabled="disabled"'
        html += ' reset-search-input="false"'
        html += ">"
        html += " <ui-select-match placeholder='{}'>".format(placeholder)
        html += ' <div ng-bind-html="$select.selected.label"></div>'
        html += " </ui-select-match>"
        html += " <ui-select-choices"
        html += '     style="{}"'.format(vstyle)
        html += '     repeat="value.id as value in options.{}"'.format(vmodel)
        html += '     refresh="{}"'.format(
            self.get_foreign(
                vurl,
                vform,
                vmodel,
                "$select.search",
                "getForeignKeys",
            ),
        )
        html += '     refresh-delay="0">'
        html += (
            '     <div ng-bind-html="value.label| highlightSelect: '
            '$select.search"></div>'
        )
        html += " </ui-select-choices>"
        html += "</ui-select>"
        return html


class MultiStaticSelect(StaticSelectMulti):
    static = True
    dynamic = False


class MultiDynamicSelect(StaticSelectMulti):
    static = False
    dynamic = True


class MultiStaticSelect_old(StaticSelect):  # noqa: N801
    multiple = True


class MultiDynamicSelect_old(DynamicSelect):  # noqa: N801
    multiple = True


class DynamicInput(DynamicSelectInputWidget, forms.widgets.Input):
    def render(self, name, value, attrs=None, choices=(), renderer=None):
        # Requet the field to render normally
        html = super().render(name, value, attrs=attrs)

        # Initialization
        vmodel = self.field_name
        vform = self.form_name
        vurl = reverse(self.autofill_url, kwargs={"search": "a"})[:-1]

        # Render
        ticket = 'uib-typeahead="item.label for item in {}"'.format(
            self.get_foreign(
                vurl,
                vform,
                vmodel,
                "$viewValue",
                "getAutoComplete",
            ),
        )
        ticket += (
            ' typeahead-on-select="autoFillFields($item, $model, '
            '$label, $event)"'
        )
        ticket += ' typeahead-wait-ms="800"'
        ticket += ' ng-change="resetAutoComplete()"'

        # Save ticket inside html
        html = html.replace('type="None"', 'type="text"')
        html = html.replace("/>", "{}/>".format(ticket))

        # Return updated html
        return html


class FileAngularInput(forms.widgets.FileInput):
    def render(self, name, value, attrs=None, renderer=None):
        if not attrs:
            attrs = {}
        if value is not None:
            attrs.update({"value": value})
        if self.is_required:
            attrs.update({"required": "required"})
        attrs.update(
            {
                "valid-file": "valid-file",
                "base-sixty-four-input": "base-sixty-four-input",
            },
        )

        # Hacer link hacia el archivo
        if value and not isinstance(value, dict):
            path_image = "{}/{}".format(settings.MEDIA_ROOT, value)
            if not os.path.exists(path_image):
                value = None
                attrs.pop("value")
                button = super().render(
                    name,
                    value,
                    attrs=attrs,
                )
                html = button
            else:
                button = super().render(
                    name,
                    value,
                    attrs=attrs,
                )
                if imghdr.what(path_image) is not None:
                    image = (
                        '<img src="{}{}" style="max-height:75px; '
                        'max-width:150px;" />'.format(
                            settings.MEDIA_URL,
                            value,
                        )
                    )
                    link = '<a href="{}{}" target="_blank">{}</a>'.format(
                        settings.MEDIA_URL,
                        value,
                        image,
                    )

                    html = '<div class="row">'
                    html += '   <div class="col-md-6">{}</div>'.format(button)
                    html += '   <div class="col-md-6">{}</div>'.format(link)
                    html += "</div>"
                else:
                    link = (
                        '<br /><a href="{}{}" '
                        'target="_blank">{}</a><br />'.format(
                            settings.MEDIA_URL,
                            value,
                            value,
                        )
                    )
                    html = link + button
        else:
            button = super().render(
                name,
                value,
                attrs=attrs,
            )
            html = button

        return html

    def value_from_datadict(self, data, files, name):
        # Get field
        field = data[name]

        if isinstance(field, dict):
            # Prepare filename
            temp_hexname = "{}{}".format(
                field["filename"].encode("ascii", "ignore"),
                random.random(),
            )
            hexname = hashlib.sha1(
                temp_hexname.encode(),
                usedforsecurity=False,
            ).hexdigest()

            ext = field["filename"].split(".")[-1]
            if not ext:
                ext = "dat"

            # Prepare temporal file in memory
            f = io.BytesIO()
            f.name = "{}.{}".format(hexname, ext)
            f.original_name = field["filename"]
            f.write(base64.b64decode(field["base64"]))
            f.seek(0)

            # Add the files to FILES
            files[name] = File(f)

            # Let Django do its work
            return super().value_from_datadict(
                data,
                files,
                name,
            )
        else:
            return {}


# class Date2TimeInput(forms.widgets.TextInput):
class Date2TimeInput(forms.widgets.DateTimeInput):
    def render(self, name, value, attrs=None, renderer=None):
        if not attrs:
            attrs = {}

        if "label" in attrs.keys():
            label = attrs.pop("label")
        else:
            label = True

        if self.is_required:
            attrs.update({"required": "required"})
        attrs.update({"class": "form-control ng-valid ng-pristine"})

        value_date = ""  # str(datetime.today().date())
        value_time = ""  # '0000'

        try:
            list_type = [
                str,
                unicode,
            ]
        except NameError:
            list_type = [
                str,
            ]

        if value and value != "":
            if isinstance(value, datetime):
                value_date = value.date()
                value_time = "{:02d}{:02d}".format(
                    value.time().hour,
                    value.time().minute,
                )
            elif type(value) in list_type:
                # 2015-06-03 00:00 or 2015-06-03 00:00:00
                try:
                    """
                    ###########################################################################
                    # WARNING: I suspect this is a bad patch because it doesn't detect the    #
                    #          format of the data depending on the Django's localization      #
                    #          system. It should checked how is being used in                 #
                    #          method value_from_datadict() from this same class              #
                    ###########################################################################
                    """  # noqa: E501
                    if len(value) == 16:
                        aux = datetime.strptime(value, "%Y-%m-%d %H:%M")
                    elif len(value) == 19:
                        aux = datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
                    else:
                        aux = None
                except ValueError:
                    aux = None

                if aux:
                    value_date = aux.date()
                    value_time = "{:02d}{:02d}".format(
                        aux.time().hour,
                        aux.time().minute,
                    )

        startview = 2
        minview = 2
        maxview = 4
        icon = "calendar"

        """
        ##############################################################################
        # WARNING: Language code should be detected on runtime from the user request #
        ##############################################################################
        """  # noqa: E501
        langcode = settings.LANGUAGE_CODE
        format_date = (
            formats.get_format("DATETIME_INPUT_FORMATS", lang=langcode)[0]
            .replace("%", "")
            .replace("d", "dd")
            .replace("m", "mm")
            .replace("Y", "yyyy")
            .replace("H", "hh")
            .replace("M", "ii")
            .split(" ")[0]
        )
        language = settings.LANGUAGE_CODE
        # language = {{LANGUAGE_CODE|default:"en"}}

        if label:
            html = (
                '<label style="margin-right:10px" '
                'class="pull-right" for="id_date">' + _("Hour") + "</label>"
            )
        else:
            html = ""
        html += '<div class="row">'
        html += '<div class="col-sm-8">'
        html += ' <div class="dropdown">'
        html += '    <div id="date_{}" class="input-group date">'.format(name)

        tmp = []
        for x in attrs:
            y = attrs[x]
            tmp.append(x + '="' + y + '"')
        attributes = " ".join(tmp)

        html += (
            '        <input type="text" name="{0}" id="id_{0}" '
            'value="{1}" {2} />'.format(name, value_date, attributes)
        )
        html += (
            '        <span class="input-group-addon">'
            '<i class="glyphicon glyphicon-{}"></i></span>'.format(icon)
        )
        html += "    </div>"
        html += " </div>"
        html += "</div>"
        html += '<div class="col-sm-4">'

        ngmodel = attrs["ng-model"].split("']")
        if len(ngmodel) > 1:
            ngmodel[0] = "{}_time".format(ngmodel[0])
            ngmodel = "']".join(ngmodel)
        else:
            ngmodel = attrs["ng-model"] + "_time"
        # attrs['ng-model'] = attrs['ng-model']+'_time'
        attrs["ng-model"] = ngmodel
        tmp = []
        for x in attrs:
            y = attrs[x]
            tmp.append(x + '="' + y + '"')
        attributes = " ".join(tmp)
        html += (
            '        <input type="text" name="{0}_time" id="id_{0}_time" '
            'value="{1}" maxlength="4" {2} />'.format(
                name,
                value_time,
                attributes,
            )
        )
        html += "</div>"
        html += "</div>"
        html += '<script type="text/javascript"> '
        html += '$("#date_{}").datetimepicker({}'.format(name, "{")
        html += ' format: "{}",'.format(format_date)
        html += " autoclose: true,"
        html += ' language:"{}",'.format(language)
        html += " todayBtn: true,"
        html += " weekStart:1,"
        html += " todayHighlight:true,"
        html += ' pickerPosition:"bottom-left",'
        html += " keyboardNavigation:false,"
        html += " startView:{},".format(startview)
        html += " minView:{},".format(minview)
        html += " maxView:{},".format(maxview)
        html += " minuteStep: 15,"
        html += "});"
        html += "</script>"
        return html

    def value_from_datadict(self, data, files, name):
        # Get field
        if name in data:
            date = data[name]
            ttime = "00:00"
            if name + "_time" in data:
                ttime = data[name + "_time"].strip()

                if ttime.isnumeric():
                    if len(ttime) == 0:
                        ttime = "00:00"
                    elif len(ttime) <= 2:
                        ttime = "00:{}".format(ttime)
                    elif len(ttime) == 3:
                        ttime = "0{}:{}".format(ttime[:1], ttime[1:])
                    else:
                        ttime = "{}:{}".format(ttime[:2], ttime[2:])

            try:
                new_date = datetime.strptime(
                    "{} {}".format(date, ttime),
                    formats.get_format(
                        "DATETIME_INPUT_FORMATS",
                        lang=get_language(),
                    )[0],
                )
            except ValueError:
                new_date = None
            # new_date = "{0} {1}".format(date, ttime)
        else:
            new_date = None
            # new_date = ''

        data[name] = new_date
        return data[name]


class WysiwygAngularRender(forms.widgets.HiddenInput):
    def render_wysiwyg(
        self,
        ngmodel,
        extraif="",
        force_editors=False,
        attrs=None,
        renderer=None,
    ):
        # Recompute ngmodel
        if re.fullmatch(r"\w+", ngmodel):
            ngmodel = "$parent.{}".format(ngmodel)
        # Compute hashkey
        hashkey = attrs.get("id", str(random.randint(0, 1000)))
        # Editors
        editors = {}
        editors["textangular"] = _("Text Angular")
        editors["quill"] = _("NG Quill")
        editors["raw"] = _("Source code")
        editors["preview"] = _("Preview")
        editors_list = []
        for keyarg in editors.keys():
            editors_list.append(
                '"{0}":{{"key":"{0}","name":"{1}"}}'.format(
                    keyarg,
                    smart_str(editors[keyarg]),
                ),
            )
        editors_json = "{" + ",".join(editors_list) + "}"

        # Detect if this field is required
        if self.is_required:
            required = " required"
        else:
            required = ""

        attributes = []
        attributes_ngquill = []
        if attrs:
            for key in attrs.keys():
                if key not in ["id", "class", "ng-model", "ng-init"]:
                    attributes.append('{}="{}"'.format(key, attrs[key]))
                    if key in ["ng-blur", "ng-change"]:
                        attributes_ngquill.append(
                            'on-content-changed="{}"'.format(attrs[key]),
                        )
                    attributes_ngquill.append(
                        '{}="{}"'.format(key, attrs[key]),
                    )
        attributes = " ".join(attributes)
        attributes_ngquill = " ".join(attributes_ngquill)
        # Get normal field
        html = ""
        html += "<div ng-init='editors_{1}={0}'></div>".format(
            editors_json,
            hashkey,
        )
        # Render wysiwyg editor's selector
        if force_editors:
            ngshow = ""
        else:
            ngshow = " ng-show='block.type==\"string\"'>"
        html += "<select ng-model='editor_{1}'{0}>".format(ngshow, hashkey)
        html += (
            '<option ng-repeat="subeditor in editors_{0}" '
            'value="{{{{subeditor.key}}}}">{{{{subeditor.name}}}}'
            "</option>".format(hashkey)
        )
        html += "</select>"

        # Render wysiwyg editors
        html += (
            '<div ng-if=\'{1}editor_{2}=="preview"\' class="form-control" '
            'ng-bind-html="{0}"></div>'.format(ngmodel, extraif, hashkey)
        )
        html += (
            '<textarea ng-model="{0}" '
            'ng-if=\'{1}editor_{4}=="raw"\' class="form-control" '
            'rows="10"{2} {3}></textarea>'.format(
                ngmodel,
                extraif,
                required,
                attributes,
                hashkey,
            )
        )
        html += (
            '<ng-quill-editor ng-model="{0}" '
            "ng-if='{1}editor_{4}==\"quill\"'{2} {3}>"
            "</ng-quill-editor>".format(
                ngmodel,
                extraif,
                required,
                attributes_ngquill,
                hashkey,
            )
        )
        html += (
            '<text-angular ng-model="{0}" '
            "ng-if='{1}editor_{4}==\"textangular\"'{2} {3}>"
            "</text-angular>".format(
                ngmodel,
                extraif,
                required,
                attributes,
                hashkey,
            )
        )
        html += (
            '<textarea ng-model="{}" '
            "ng-if='{}editor' style='background-color:#fdd'{} {}>"
            "</textarea>".format(ngmodel, extraif, required, attributes)
        )

        return html


class WysiwygAngularInput(WysiwygAngularRender):
    def render(self, name, value, attrs=None, renderer=None):
        # Compute hashkey
        hashkey = attrs.get("id", str(random.randint(0, 1000)))
        # Get model name
        vmodel = attrs.get("ng-model")  # .replace("'",'"')
        init = attrs.get("ng-init", "")

        if value is None:
            value = ""
        # Render
        html = "<div ng-init='editor_{}=\"textangular\"'>".format(hashkey)
        html += (
            '<textarea name="{0}" ng-model="{1}" ng-show=\'false\' '
            'ng-init="{3}">{2}</textarea>'.format(name, vmodel, value, init)
        )
        html += self.render_wysiwyg(
            ngmodel=vmodel,
            force_editors=True,
            attrs=attrs,
            renderer=renderer,
        )
        html += "</div>"

        # Return result
        return html


class MultiBlockWysiwygInput(WysiwygAngularRender):
    def render(self, name, value, attrs=None, renderer=None):
        # Compute hashkey
        hashkey = attrs.get("id", str(random.randint(0, 1000)))

        # Cleaned value
        if value:
            value_clean = value.replace("'", "&#39;")
        else:
            value_clean = value

        # Get model name
        vmodel = attrs.get("ng-model").replace("'", '"')
        # Get normal field
        html = (
            "<div ng-init='{0}={{}}; {0}[\"__JSON_DATA__\"]={1}'>"
            "</div>".format(vmodel, value_clean)
        )
        html += "<input type='hidden' name='{}' ng-model='{}'>".format(
            name,
            vmodel,
        )

        # Render blocks with ANGULAR
        html += (
            "<div ng-repeat='(key, block) in "
            '{}["__JSON_DATA__"]\' ng-init=\'editor_{}="quill"\'>'.format(
                vmodel,
                hashkey,
            )
        )
        html += "<label>{{key}}</label>"

        html += (
            '<div ng-show="block.deleted"><p class="text-danger">{}</p>'
            "</div>".format(_("Field deleted in the template"))
        )
        html += self.render_wysiwyg(
            ngmodel="block.value",
            extraif='block.type=="string" && ',
            attrs=attrs,
            renderer=renderer,
        )
        html += "</div>"

        # Return result
        return html


class VisualHTMLInput(forms.widgets.HiddenInput):
    """
    Example
    service_day_html = forms.CharField(
        label=_("Quota"),  # Optional
        required=False,
        widget=VisualHTMLInput(),
        initial={'selfname': 'service_day_html', 'data': "<div id='<#id#>'
                 ng-init='<#form:service_day_color#>=\"#ff00ff\"'
                 color-contrast='{{<#form:service_day_color#>}}'>
                 {{<#form:service_day_color#>}}</div>"}
    )
    """

    visual = True

    def render(self, name, value, attrs=None, renderer=None):
        # Compute hashkey
        hashkey = attrs.get("id", str(random.randint(0, 1000)))
        vmodel = attrs.get("ng-model").replace("'", '"')
        selfname = self.attrs.get("selfname", None)
        value = self.attrs.get("data", None)

        # Process all tags
        if value is not None:
            for keydirty in value.split("<#")[1:]:
                key = keydirty.split("#>")[0]
                keysp = key.split(":")
                actor = keysp[0]
                args = keysp[1:]
                if actor == "form":
                    if selfname:
                        newname = vmodel.replace(selfname, args[0])
                        value = value.replace("<#{}#>".format(key), newname)
                    else:
                        raise OSError(
                            "selfname must be included in the attrs from "
                            "the widget with the name of the field in "
                            "the form",
                        )
                elif actor == "id":
                    value = value.replace("<#id#>", hashkey)
                elif actor == "ngmodel":
                    value = value.replace("<#ngmodel#>", vmodel)
        else:
            raise OSError(
                "data must be included in the attrs from the widget "
                "with the name of the field in the form",
            )

        # Return result
        return value


class BootstrapWysiwygInput(forms.widgets.HiddenInput):
    def render(self, name, value, attrs=None, renderer=None):
        # Get model name
        # vmodel = attrs.get('ng-model').replace("'",'"')

        html = _(
            "We are working on this plugin, it will be ready for next version",
        )

        # Return result
        return html


class GenReCaptchaInput(ReCaptcha):
    # Legacy is added to make the Widget to work as it was designed
    # this field should be True when is being analized to answer an API request
    # and json_details is set to True on the query from the remote client
    # The rest of the time legacy should be kept False
    legacy = False

    def __init__(self, *args, **kwargs):
        # Decide name of the field in the POST
        if "fieldname" in kwargs:
            fieldname = kwargs.pop("fieldname")
        if not fieldname:
            fieldname = "g-recaptcha-response"
        self.recaptcha_response_name = fieldname
        self.recaptcha_challenge_name = fieldname

        # Keep going as usually
        return super().__init__(*args, **kwargs)

    #    def __init__(self, public_key=None, use_ssl=None, attrs={}, *args,
    #        **kwargs):
    #        self.public_key = public_key or settings.RECAPTCHA_PUBLIC_KEY
    #        self.use_ssl = use_ssl if use_ssl is not None else
    #        getattr(settings, 'RECAPTCHA_USE_SSL', False)
    #        self.js_attrs = attrs
    #        super(ReCaptcha, self).__init__(*args, **kwargs)

    #    def render(self, name, value, attrs=None, renderer=None):
    #        return mark_safe(u'%s' % client.displayhtml(self.public_key,
    #        self.js_attrs, use_ssl=self.use_ssl))

    def render(self, name, value, attrs=None, renderer=None):
        if self.legacy:
            html = super().render(name, value, attrs)
        else:
            html = (
                '<input ng-model="{0}" name="{0}" id="id_{0}" '
                'type="hidden" ng-required="true">'.format(name)
            )
            html += (
                '<div vc-recaptcha key="\'{}\'" ng-model="{}"></div>'.format(
                    self.public_key,
                    name,
                )
            )
        return html



from django.contrib.auth.decorators import user_passes_test
from django.core.exceptions import PermissionDenied


def restrict_access_to_groups(groups: list, raise_exception: bool = False):
    """
    Decorator for views that requires the user to be part of a group,
    if they are not the user is not allowed into the page.
    If the raise_exception parameter is given the PermissionDenied exception
    is raised returning a 403 status code

    Usage:

    @method_decorator(
        restrict_access_to_groups(
            ["admin", "sales"],
            True
        ),
        name="dispatch"
    )

    Author: Ben Cleary (https://github.com/bencleary)
    """

    def in_groups(user):
        # checks if the user is authenticated, if not returns False
        if not user.is_authenticated:
            return False

        # checks if the user is a superuser or is part of the given groups
        if user.groups.filter(name__in=groups).exists() | user.is_superuser:
            return True

        # if raise_exception is given raise the 403 error
        if raise_exception:
            raise PermissionDenied

        # return False otherwise
        return False

    return user_passes_test(in_groups)



from django.db import models
from multi_email_field.fields import MultiEmailField


class TestModel(models.Model):
    f = MultiEmailField(null=True, blank=True)
    f_default = MultiEmailField(
        default=["test@example.com"],
        null=True,
        blank=True,
    )



# Generated by Django 1.10.5 on 2017-01-18 17:42
# Simplified by Juanmi Taboada on 2017-14-02 11:12

from typing import Any, List

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("codenerix", "0006_auto_20160825_0838"),
    ]

    operations: List[Any] = []



#
# django-codenerix
#
# Codenerix GNU
#
# Project URL : http://www.codenerix.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from bson import json_util
from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import models
from django.db.models import Q
from django.db.models.signals import post_delete, pre_delete
from django.dispatch.dispatcher import receiver
from django.utils.encoding import force_str, smart_str
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta

from codenerix.helpers import daterange_filter
from codenerix.middleware import get_current_user

# Separator to log
SEPARATOR = "\u8594"
SEPARATOR_HTML = " &rarr; "


class CodenerixMetaType(dict):
    """
    Define type for CodenerixMeta of the instance NOT the class
    Example:
    m = CodenerixMetaType({'first_name': 'Eduardo'}, last_name='Pool', age=24,
    sports=['Soccer'])
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k in arg:
                    self[k] = arg[k]

        if kwargs:
            for k in kwargs:
                self[k] = kwargs[k]

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super().__delitem__(key)
        del self.__dict__[key]

    def __getnewargs__(self):
        return tuple()

    def __getstate__(self):
        return self.__dict__


class CodenerixModelBase(models.Model):
    class Meta(TypedModelMeta):
        abstract = True

    # return method relation objects
    def __getmro__(self):
        return self.__class__.__mro__

    # recolecta informacion de todas las clases que intervienen en la instancia
    # collects information from all classes that intervene in the instance
    def __init__(self, *args, **kwards):
        self.CodenerixMeta = CodenerixMetaType()

        mro = self.__getmro__()
        for cl in reversed(mro):
            if "CodenerixMeta" in cl.__dict__.keys():
                for key in cl.CodenerixMeta.__dict__.keys():
                    if "__" != key[0:2]:
                        value = getattr(cl.CodenerixMeta, key)
                        if value:
                            if key not in self.CodenerixMeta:
                                self.CodenerixMeta[key] = value
                            else:
                                if isinstance(value, dict):
                                    self.CodenerixMeta[key].update(value)
                                elif isinstance(value, list):
                                    if not isinstance(
                                        self.CodenerixMeta[key],
                                        list,
                                    ):
                                        self.CodenerixMeta[key] = list(
                                            self.CodenerixMeta[key],
                                        )
                                    self.CodenerixMeta[key] += value
                                elif isinstance(self.CodenerixMeta[key], list):
                                    self.CodenerixMeta[key] += list(value)

        return super().__init__(*args, **kwards)


class CodenerixModel(CodenerixModelBase):
    """
    Special methods are
        __fields__: it is a list of fields
            Usage:      fields.append(('key','Name',size:int_in_pixels,'alignment:left|right|center'))
            Example 1:  fields.append(('title',_('Title')))
            Example 2:  fields.append(('title',_('Title'),100,'center'))
            Example 3:  fields.append(('title',_('Title'),None,'center'))   # We don't want to define the size but we want to define the alignment
            Example 4:  fields.append((None,_('Title')))                    # We don't want any ordering in the field Title
            Example 5:  fields.append(('user__username',_('Username')))     # You can define here relationships as well

        __limitQ__:
        __searchQ__:
        __searchF__:
    """  # noqa: E501

    created = models.DateTimeField(  # type: ignore[var-annotated]
        _("Created"),
        editable=False,
        auto_now_add=True,
    )
    updated = models.DateTimeField(  # type: ignore[var-annotated]
        _("Updated"),
        editable=False,
        auto_now=True,
    )

    class Meta(TypedModelMeta):
        abstract = True
        default_permissions = ("add", "change", "delete", "view", "list")

    class CodenerixMeta:
        abstract = None

    def __init__(self, *args, **kwards):
        self.CodenerixMeta = CodenerixMetaType()
        self.codenerix_uuid = None
        self.codenerix_request = None
        return super().__init__(*args, **kwards)

    def __strlog_add__(self):
        return ""

    def __strlog_update__(self, newobj):
        return ""

    def __strlog_delete__(self):
        return ""

    def __limitQ__(self, info):  # noqa: N802
        return {}

    def __searchQ__(self, info, text):  # noqa: N802
        return {}

    def __searchF__(self, info):  # noqa: N802
        return {}

    def lock_update(self, request=None):
        return None

    @property
    def codenerix_uuid(self):
        return self.__codenerix_uuid

    @codenerix_uuid.setter
    def codenerix_uuid(self, uuid):
        self.__codenerix_uuid = uuid
        return uuid

    @property
    def codenerix_request(self):
        return self.__codenerix_request

    @codenerix_request.setter
    def codenerix_request(self, request):
        self.__codenerix_request = request
        return request

    def internal_lock_delete(self):
        # if we have a specific lock delete from model
        answer = self.lock_delete()
        if answer is None:
            # for each field
            for related in self._meta.get_fields():
                # check if it is protected
                if (
                    "on_delete" in related.__dict__
                    and related.on_delete == models.PROTECT
                ):
                    # if we have a name
                    field = getattr(self, related.related_name, None)
                    if field:
                        # try to get 'exists' function
                        f_exists = getattr(field, "exists", None)
                        # if we didn't get it or the result from it is positive
                        if f_exists is None or f_exists():
                            # answer that the item is locked
                            answer = _(
                                "Cannot delete item, relationship "
                                "with %(model_name)s",
                            ) % {
                                "model_name": related.related_model._meta.verbose_name,  # noqa: E501
                            }
                            break
        return answer

    def lock_delete(self, request=None):
        return None

    # check lock update
    def clean(self):
        locked = self.lock_update()
        if locked is not None:
            raise ValidationError(locked)
        else:
            return super().clean()


class GenInterface(CodenerixModelBase):
    """
    Check force_methods options in CodenerixMeta class and it makes sure that
    the specified methods exists
    """

    class Meta(TypedModelMeta):
        abstract = True

    class CodenerixMeta:
        """
        force_methods = {'alias': ('method_name', 'Description'), }
        """

        pass

    def __init__(self, *args, **kwards):
        self.CodenerixMeta = CodenerixMetaType()
        super().__init__(*args, **kwards)

        # revisamos que esten implementados los metodos indicados
        # we checked that the indicated methods are implemented
        force_methods = getattr(self.CodenerixMeta, "force_methods", None)
        if force_methods:
            for alias in force_methods.keys():
                method = force_methods[alias]
                if not hasattr(self, method[0]) or not callable(
                    getattr(self, method[0]),
                ):
                    raise OSError(
                        "Method {}() not found in class {}: {}".format(
                            method[0],
                            self._meta.object_name,
                            method[1],
                        ),
                    )
        return super().__init__(*args, **kwards)


@receiver(pre_delete)
def codenerixmodel_delete_pre(sender, instance, **kwargs):
    if (
        not hasattr(instance, "name_models_list")
        and hasattr(instance, "internal_lock_delete")
        and callable(instance.internal_lock_delete)
    ):
        lock_delete = instance.internal_lock_delete()
        if lock_delete is not None:
            raise PermissionDenied(lock_delete)


# We don't use log system when PQPRO_CASSANDRA == TRUE
if not (hasattr(settings, "PQPRO_CASSANDRA") and settings.PQPRO_CASSANDRA):  # type: ignore[misc] # noqa: E501
    from django.contrib.admin.models import ADDITION, CHANGE, DELETION

    TYPE_ACTION = (
        (ADDITION, _("Add")),
        (CHANGE, _("Change")),
        (DELETION, _("Delete")),
    )

    class Log(models.Model):
        """
        Control the possible log
        """

        action_time = models.DateTimeField(  # type: ignore[var-annotated]
            "Date",
            auto_now=True,
        )
        user = models.ForeignKey(  # type: ignore[var-annotated]
            settings.AUTH_USER_MODEL,
            on_delete=models.SET_NULL,
            blank=True,
            null=True,
            related_name="log",
        )
        username = models.CharField(  # type: ignore[var-annotated]
            "Username",
            max_length=200,
            blank=True,
            null=False,
            default="",
        )
        content_type = models.ForeignKey(  # type: ignore[var-annotated]
            ContentType,
            on_delete=models.SET_NULL,
            blank=True,
            null=True,
            related_name="log",
        )
        object_id = models.TextField(  # type: ignore[var-annotated]
            "Object id",
            blank=True,
            null=True,
        )
        object_repr = models.CharField(  # type: ignore[var-annotated]
            "Object repr",
            max_length=200,
        )
        action_flag = models.PositiveSmallIntegerField(  # type: ignore[var-annotated] # noqa: E501
            _("Action"),
            choices=TYPE_ACTION,
        )
        change_json = models.TextField(  # type: ignore[var-annotated]
            "Json",
            blank=True,
            null=False,
        )
        change_txt = models.TextField(  # type: ignore[var-annotated]
            "Txt",
            blank=True,
            null=False,
        )
        snapshot_txt = models.TextField(  # type: ignore[var-annotated]
            "Snapshot Txt",
            blank=True,
            null=False,
        )

        class Meta(TypedModelMeta):
            permissions = [
                ("list_log", "Can list log"),
                ("detail_log", "Can view log"),
            ]

        def show(self, view="html"):
            text = []
            if self.change_txt:
                cambios = json.loads(self.change_txt)
            else:
                cambios = ""
            for c in cambios:
                if isinstance(cambios[c], list):
                    text.append("{}: {}".format(cambios[c][0], cambios[c][1]))
                else:
                    text.append("{}: {}".format(_(c), cambios[c]))

            if view == "html":
                result = mark_safe(
                    "<ul><li>{}</li></ul>".format(
                        "</li><li>".join(text),
                    ).replace(SEPARATOR, SEPARATOR_HTML),
                )
            else:
                result = "\n".join(text)

            return result

        def __unicode__(self):
            return self.show(view="txt")

        def __str__(self):
            return self.__unicode__()

        def action(self):
            # Find the action
            if self.action_flag == ADDITION:
                answer = _("Add")
            elif self.action_flag == CHANGE:
                answer = _("Change")
            elif self.action_flag == DELETION:
                answer = _("Delete")
            else:
                answer = "?"
            # Return answer
            return answer

        def __fields__(self, info):
            fields = []
            fields.append(("action_time", _("Date")))
            fields.append(("user__username", _("Actual user")))
            fields.append(("username", _("Original user")))
            # fields.append(('content_type__name', _('APP Name')))
            fields.append(("content_type", _("APP Name")))
            fields.append(("get_action_flag_display", _("Action")))
            # fields.append(("content_type__app_label", _("APP Label")))
            # fields.append(("content_type__model", _("APP Model")))
            fields.append(("object_id", _("ID")))
            # fields.append(('object_repr', _('Representation')))
            fields.append(("show", _("Txt")))
            return fields

        def __searchQ__(self, info, text):  # noqa: N802
            tf = {}
            tf["user"] = Q(user__username__icontains=text)
            tf["username"] = Q(username__icontains=text)
            # Flag
            if (
                (text.lower() == "add")
                or (text.lower() == "addition")
                or (text.lower() == _("add"))
                or (text.lower() == _("addition"))
            ):
                tf["action_flag"] = Q(action_flag=ADDITION)
            elif (
                (text.lower() == "change")
                or (text.lower() == "changed")
                or (text.lower() == _("change"))
                or (text.lower() == _("changed"))
            ):
                tf["action_flag"] = Q(action_flag=CHANGE)
            elif (
                (text.lower() == "edit")
                or (text.lower() == "edition")
                or (text.lower() == _("edit"))
                or (text.lower() == _("edition"))
            ):
                tf["action_flag"] = Q(action_flag=CHANGE)
            elif (
                (text.lower() == "delete")
                or (text.lower() == "deleted")
                or (text.lower() == _("delete"))
                or (text.lower() == _("deleted"))
            ):
                tf["action_flag"] = Q(action_flag=DELETION)
            # tf['content_type'] = Q(content_type__name__icontains=text)
            tf["object_id"] = Q(object_id__icontains=text)
            tf["object_repr"] = Q(object_repr__icontains=text)
            tf["action_time"] = "datetime"
            return tf

        def __searchF__(self, info):  # noqa: N802
            tf = {}
            tf["action_time"] = (
                _("Date"),
                lambda x: Q(**daterange_filter(x, "action_time")),
                "daterange",
            )
            tf["get_action_flag_display"] = (
                _("Action"),
                lambda x: Q(action_flag=x),
                list(TYPE_ACTION),
            )
            tf["object_id"] = (_("ID"), lambda x: Q(object_id=x), "input")
            tf["user__username"] = (
                _("Actual user"),
                lambda x: Q(user__username__icontains=x),
                "input",
            )
            tf["username"] = (
                _("Original user"),
                lambda x: Q(username__icontains=x),
                "input",
            )
            tf["content_type"] = (
                _("APP Name"),
                lambda x: Q(content_type__pk=x),
                sorted(
                    [
                        (obj["pk"], f"{obj['app_label']} | {obj['model']}")
                        for obj in ContentType.objects.filter(
                            log__isnull=False,
                        )
                        .all()
                        .values("pk", "app_label", "model")
                        .distinct()
                    ],
                    key=lambda x: x[1].lower(),
                ),
            )
            # tf["content_type__app_label"] = (
            #     _("APP Label"),
            #     lambda x: Q(content_type__app_label__icontains=x),
            #     "input",
            # )
            # tf["content_type__model"] = (
            #     _("APP Model"),
            #     lambda x: Q(content_type__model__icontains=x),
            #     "input",
            # )
            tf["show"] = (
                _("Txt"),
                lambda x: Q(change_txt__icontains=x),
                "input",
            )
            # tf['users']=(
            #   _('User'),
            #   lambda x: Q(user__username=x),[('M','M*'),('S','S*')]
            # )
            return tf

    class GenLog:
        class CodenerixMeta(CodenerixModel.CodenerixMeta):
            log_full = False

        def post_save(self, log):
            # custom post save from application
            pass

        def save(self, *args, **kwargs):
            user = get_current_user()
            if user:
                user_id = user.pk
                username = user.username
            else:
                user_id = None
                username = ""

            model = apps.get_model(
                self._meta.app_label,
                self.__class__.__name__,
            )
            isnew = True
            if self.pk is not None:
                list_obj = model.objects.filter(pk=self.pk)
                isnew = list_obj.count() == 0
                # raise IOError,self.__dict__
            # only attributes changes
            attrs = {}
            attrs_txt = {}
            # attributes from database
            attrs_bd = {}
            if isnew:
                action = ADDITION
                pk = None
            else:
                action = CHANGE
                pk = self.pk
                # Instance object
                # obj = model.objects.get(pk=self.pk)
                obj = list_obj.get()
                for key in obj._meta.get_fields():
                    key = key.name
                    # exclude manytomany
                    if obj._meta.model._meta.local_many_to_many and key in [
                        x.name
                        for x in obj._meta.model._meta.local_many_to_many
                    ]:
                        value = None
                    elif obj._meta.get_fields(include_hidden=True) and key in [
                        x.name
                        for x in obj._meta.get_fields(include_hidden=True)
                        if x.many_to_many and x.auto_created
                    ]:
                        value = None
                    else:
                        value = getattr(obj, key, None)
                    attrs_bd[key] = value

            # comparison attributes
            # for key in self._meta.get_fields():
            aux = None
            list_fields = [x.name for x in self._meta.get_fields()]
            for ffield in self._meta.get_fields():
                key = ffield.name
                # exclude manytomany
                if self._meta.model._meta.local_many_to_many and key in [
                    x.name for x in self._meta.model._meta.local_many_to_many
                ]:
                    field = None
                # elif self._meta.get_all_related_many_to_many_objects() and key in [x.name for x in self._meta.get_all_related_many_to_many_objects()]: # noqa: E501
                elif self._meta.get_fields(include_hidden=True) and key in [
                    x.name
                    for x in self._meta.get_fields(include_hidden=True)
                    if x.many_to_many and x.auto_created
                ]:
                    field = None
                else:
                    field = getattr(self, key, None)

                if key in list_fields:
                    # if (not attrs_bd.has_key(key))
                    # or (field != attrs_bd[key]):
                    if (key not in attrs_bd) or (field != attrs_bd[key]):
                        if field is not None or action == CHANGE:
                            aux = ffield
                            field_txt = field
                            if field_txt is None:
                                field_txt = "---"
                            if isinstance(field, CodenerixModel):
                                field = field.pk

                            try:
                                json.dumps(field, default=json_util.default)
                                if (
                                    key not in attrs_bd
                                    or not self.CodenerixMeta.log_full
                                ):
                                    attrs[key] = field
                                else:
                                    if isinstance(
                                        attrs_bd[key],
                                        CodenerixModel,
                                    ):
                                        attrs[key] = (
                                            attrs_bd[key].pk,
                                            field,
                                        )
                                    else:
                                        attrs[key] = (
                                            attrs_bd[key],
                                            field,
                                        )
                            except Exception:
                                # If related, we don't do anything
                                if getattr(field, "all", None) is None:
                                    field = str(field)
                                    if (
                                        key not in attrs_bd
                                        or not self.CodenerixMeta.log_full
                                    ):
                                        attrs[key] = field
                                    else:
                                        attrs[key] = (
                                            attrs_bd[key],
                                            field,
                                        )

                            if hasattr(ffield, "verbose_name"):
                                try:
                                    is_string = isinstance(
                                        ffield.verbose_name,
                                        unicode,
                                    ) or isinstance(ffield.verbose_name, str)

                                except NameError:
                                    is_string = isinstance(
                                        ffield.verbose_name,
                                        str,
                                    )

                                if is_string:
                                    ffield_verbose_name = ffield.verbose_name
                                else:
                                    ffield_verbose_name = str(
                                        ffield.verbose_name,
                                    )

                                if (
                                    key not in attrs_bd
                                    or not self.CodenerixMeta.log_full
                                ):
                                    attrs_txt[ffield_verbose_name] = force_str(
                                        field_txt,
                                        errors="replace",
                                    )
                                else:
                                    if attrs_bd[key] is None:
                                        attrs_bd[key] = "---"
                                    attrs_txt[key] = (
                                        ffield_verbose_name,
                                        "{}{}{}".format(
                                            force_str(
                                                attrs_bd[key],
                                                errors="replace",
                                            ),
                                            SEPARATOR,
                                            force_str(
                                                field_txt,
                                                errors="replace",
                                            ),
                                        ),
                                    )

            log = Log()
            log.user_id = user_id
            log.username = username
            log.content_type_id = ContentType.objects.get_for_model(self).pk
            log.object_id = pk
            log.object_repr = force_str(self, errors="replace")[:200]
            try:
                log.change_json = json.dumps(attrs, default=json_util.default)
            except UnicodeDecodeError:
                log.change_json = json.dumps(
                    {"error": "*JSON_ENCODE_ERROR*"},
                    default=json_util.default,
                )
            try:
                log.change_txt = json.dumps(
                    attrs_txt,
                    default=json_util.default,
                )
            except UnicodeDecodeError:
                log.change_txt = json.dumps(
                    {"error": "*JSON_ENCODE_ERROR*"},
                    default=json_util.default,
                )
            log.action_flag = action
            if pk is None:
                log.snapshot_txt = self.__strlog_add__()
            else:
                log.snapshot_txt = obj.__strlog_update__(self)

            aux = super().save(*args, **kwargs)
            if pk is None:
                # if new element, get pk
                log.object_id = self.pk
            log.save()

            # custom post save from application
            self.post_save(log)

            return aux

    class GenLogFull(GenLog):
        class CodenerixMeta(CodenerixModel.CodenerixMeta):
            log_full = True

    @receiver(post_delete)
    def codenerixmodel_delete_post(sender, instance, **kwargs):
        if not hasattr(instance, "name_models_list") and issubclass(
            sender,
            GenLog,
        ):
            user = get_current_user()
            if user:
                user_id = user.pk
                username = user.username
            else:
                user_id = None
                username = "*Unknown*"
            action = DELETION

            attrs = {}
            attrs_txt = {}

            # ._meta.get_fields() return all fields include related name
            # ._meta.fields return all fields of models

            # list_fields = [x.name for x in instance._meta.get_fields()]
            for ffield in instance._meta.get_fields():
                key = ffield.name
                # exclude manytomany
                if instance._meta.model._meta.local_many_to_many and key in [
                    x.name
                    for x in instance._meta.model._meta.local_many_to_many
                ]:
                    field = None
                # elif self._meta.get_all_related_many_to_many_objects() and key in [x.name for x in self._meta.get_all_related_many_to_many_objects()]:  # noqa: E501
                elif instance._meta.get_fields(
                    include_hidden=True,
                ) and key in [
                    x.name
                    for x in instance._meta.get_fields(include_hidden=True)
                    if x.many_to_many and x.auto_created
                ]:
                    field = None
                else:
                    try:
                        field = getattr(instance, key, None)
                    except Exception:
                        field = None

                # If we have information to register
                if field is not None:
                    if isinstance(field, CodenerixModel):
                        attrs[key] = field.pk
                    else:
                        try:
                            json.dump(field, default=json_util.default)
                            attrs[key] = field
                        except TypeError:
                            # If related, we don't do anything
                            if getattr(field, "all", None) is None:
                                # field_str = str(field)
                                field_str = smart_str(field)
                                attrs[key] = field_str

                    if hasattr(ffield, "verbose_name"):
                        try:
                            is_string = isinstance(
                                ffield.verbose_name,
                                unicode,
                            ) or isinstance(ffield.verbose_name, str)

                        except NameError:
                            is_string = isinstance(ffield.verbose_name, str)

                        if is_string:
                            ffield_verbose_name = ffield.verbose_name
                        else:
                            ffield_verbose_name = str(ffield.verbose_name)

                        attrs_txt[ffield_verbose_name] = force_str(
                            field,
                            errors="replace",
                        )

            try:
                representation = force_str(instance)[:200]
            except Exception:
                representation = "*Unknown*"

            log = Log()
            log.user_id = user_id
            log.username = username
            log.content_type_id = ContentType.objects.get_for_model(
                instance,
            ).pk
            log.object_id = instance.pk
            log.object_repr = representation
            log.change_json = json.dumps(attrs, default=json_util.default)
            log.change_txt = json.dumps(attrs_txt, default=json_util.default)
            log.snapshot_txt = instance.__strlog_delete__()
            log.action_flag = action
            log.save()

    class RemoteLog(CodenerixModel):
        """
        RemoteLog system
        """

        user = models.ForeignKey(  # type: ignore[var-annotated]
            settings.AUTH_USER_MODEL,
            on_delete=models.SET_NULL,
            blank=True,
            null=True,
            related_name="remotelog",
        )
        username = models.CharField(  # type: ignore[var-annotated]
            "Username",
            max_length=200,
            blank=True,
            null=False,
            default="",
        )
        data = models.TextField(  # type: ignore[var-annotated]
            "Data",
            blank=False,
            null=False,
        )

        def __fields__(self, info):
            fields = []
            fields.append(("pk", _("ID")))
            fields.append(("created", _("Created")))
            fields.append(("user", _("Actual user")))
            fields.append(("username", _("Original user")))
            return fields

        def save(self, *args, **kwargs):
            if self.user:
                self.username = self.user.username
            super().__init__(*args, **kwargs)



import os

from setuptools import setup

import codenerix

# from setuptools.command.install import install
# class CustomInstallCommand(install):
#    """Customized setuptools install command - prints a friendly greeting."""
#    def run(self):
#        print "Hello, developer, how are you? :)"
#        install.run(self)
#        #post-processing code

with open(os.path.join(os.path.dirname(__file__), "README.rst")) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="django_codenerix",
    version=codenerix.__version__,
    packages=["codenerix"],
    include_package_data=True,
    zip_safe=False,
    license="Apache License Version 2.0",
    description="Codenerix it is a framework that goes on top of Django so "
    "it makes easier development and building of ERPs.",
    long_description=README,
    url="https://github.com/codenerix/django-codenerix",
    author=", ".join(codenerix.__authors__),
    author_email=", ".join(codenerix.__authors_email__),
    keywords=["django", "codenerix", "management", "erp", "crm"],
    platforms=["OS Independent"],
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 4.0",
        "Intended Audience :: Developers",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    # cmdclass={ 'install': CustomInstallCommand, },
    install_requires=[
        "codenerix_lib>=1.0.27",
        "pymongo",
        "python-dateutil",
        "django-recaptcha>=1.2.1,<1.3",
        "django-rosetta>=0.9.8",
        "jsonfield",
        "openpyxl==3.1.2",
        "Pillow",
        "Unidecode",
        "Django>=4.1.1",
        "ldap3",
        "django-haystack>=2.6.1",
        "pytz",
        "elasticsearch>=2.0.0,<3.0.0",
        "django-debug-toolbar>=4.4.2",
        "django-debug-toolbar-request-history",
        "pyotp",
        "html5lib",
        "pyinstrument",
        "django-stubs",
        "django-stubs-ext",
    ],
)



#
# django-codenerix
#
# Codenerix GNU
#
# Project URL : http://www.codenerix.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

try:
    from subprocess import getstatusoutput

    pythoncmd = "python3"
except Exception:
    from commands import (  # type: ignore[import-not-found,no-redef]
        getstatusoutput,
    )

    pythoncmd = "python2"

from codenerix_lib.colors import colors
from codenerix_lib.debugger import Debugger
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand, Debugger):
    # Show this when the user types help
    help = "Show colors for Debugger"

    def handle(self, *args, **options):
        # Autoconfigure Debugger
        self.set_name("CODENERIX")
        self.set_debug()

        # Reorder colors
        keys = []
        for key in colors.keys():
            keys.append((colors[key][0], colors[key][1], key))
        keys.sort()

        # Show up all colors
        for color in keys:
            # Get the color information
            (simplebit, subcolor) = colors[color[2]]
            # Show it
            print(
                "{:1d}:{:02d} - \033[{:1d};{:02d}m{:<14s}\033[1;00m{:<s}".format(  # noqa: E501
                    simplebit,
                    subcolor,
                    simplebit,
                    subcolor,
                    color[2],
                    color[2],
                ),
            )

            # Get environment
            appname = settings.ROOT_URLCONF.split(".")[0]
            basedir = settings.BASE_DIR
            appdir = os.path.abspath("{}/{}".format(basedir, appname))
            status, output = getstatusoutput(
                "find {}/ -type f -name '*.pyc' -delete".format(appdir),
            )
            if status:
                raise CommandError(output)
